/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005-6 Aidan Thornton <makomk@lycos.co.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <stdlib.h>
#include "ngcs_common.h"
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>


/*! \brief Sends a ngcs message on the specified socket
 *
 *  This function is used to send a raw ngcs message on a socket. The message
 *  is expected to conform to certain conventions, but these are not checked.
 *  (See the definitions of the NGCS_TYPE constants for details of exactly what
 *  is expected with each one).
 * 
 *  \param sock the socket on which the message is sent
 *  \param chan the channel number the message is marked with
 *  \param type the data type of the message (typically one of the NGCS_TYPE
 *         constants)
 *  \param len the length of the message body (in bytes)
 *  \param data the message body
 *  \return Zero on success, non-zero on failure
 *  \sa ngcs_recvmsg(), ngcs_pack()
 */
int ngcs_sendmsg(int sock, int chan, int type, int len, char *data)
{
    int head[3];

    if (len > 0)
        assert(data);
    head[0] = chan;
    head[1] = type;
    head[2] = len;
    if (ngcs_sendall(sock, head, 3 * sizeof(int)))
        return 1;
    if (len <= 0)
        return 0;
    if (ngcs_sendall(sock, data, len))
        return 1;
    return 0;
}

/*! \brief Receives a raw ngcs message from the specified socket
 *
 * This function reads one raw ngcs message from the specified socket,
 * allocating the neccesary memory to hold its contents (which must be 
 * free()d by the caller)
 *
 * \param sock the socket from which the message should be read
 * \param chan a pointer to a location where the channel number of the message
 *        should be stored
 * \param type a pointer to a location where the data type of the message
 *        should be stored
 * \param len a pointer to a location where the length of the message body (in bytes)
 *        should be stored. On return, *len may be less than zero, in which case 
 *        *data is set to NULL
 * \param data a location where a pointer to the message body is stored. *data is
 *        valid on return if ngcs_recvmsg() succeeded with *len>=0, and in this case
 *        must be free()d by the caller after use.
 * \return Zero on success, non-zero on failure
 */
int ngcs_recvmsg(int sock, int *chan, int *type, int *len, char **data)
{
    int head[3];

    assert(chan);
    assert(type);
    assert(len);
    assert(data);

    if (ngcs_recvall(sock, head, 3 * sizeof(int)))
        return 1;
    *chan = head[0];
    *type = head[1];
    *len = head[2];
    if (*len < 0)
    {
        *data = NULL;
        return 0;
    }
    *data = malloc(*len);
    if (*data == NULL)
        return 1;
    if (ngcs_recvall(sock, *data, *len))
        return 1;
    return 0;
}


/*! \brief Pack an ngcs stucture for transmission
 * 
 * This function packs a sequence of items into a buffer, which can be transmitted
 * down a socket or otherwise transferred, then unpacked by ngcs_unpack()
 * The resulting packed data strcuture should be transmitted marked as type
 * NGCS_TYPE_STRUCT.
 *
 * Since the caller must ensure that the buffer is large enough, your code should
 * call ngcs_pack with buf=NULL to get the required size, then allocate a buffer
 * and call it again with the same data an the newly-allocated buffer
 *
 * \param data an array of structures describing the items of data to pack
 * \param cnt the number of items to pack (the length of data[])
 * \param buf the buffer into which to pack the data. May be null, in which case
 *            the required buffer size is returned
 *
 * \return the total packed length of the data, in bytes
 */
int ngcs_pack(ngcs_data * data, int cnt, char *buf)
{
    int n;
    int outcnt = 0;

    for (n = 0; n < cnt; n++)
    {
        if (buf)
        {
            *(int *) buf = data[n].type;
            buf += sizeof(int);
        }
        switch (data[n].type)
        {
            case NGCS_TYPE_INT:
                outcnt += 3 * sizeof(int);
                if (buf)
                {
                    *(int *) buf = sizeof(int);
                    buf += sizeof(int);
                    *(int *) buf = data[n].d.i;
                    buf += sizeof(int);
                }
                break;
            case NGCS_TYPE_LONG:
                outcnt += 2 * sizeof(int) + sizeof(long);
                if (buf)
                {
                    *(int *) buf = sizeof(long);
                    buf += sizeof(int);
                    *(int *) buf = data[n].d.l;
                    buf += sizeof(long);
                }
                break;
            case NGCS_TYPE_STRING:
                if (data[n].len < 0)
                    data[n].len = strlen(data[n].d.s);
                outcnt += 2 * sizeof(int) + data[n].len;
                if (buf)
                {
                    *(int *) buf = data[n].len;
                    buf += sizeof(int);
                    if (data[n].len)
                        memcpy(buf, data[n].d.s, data[n].len);
                    buf += data[n].len;
                }
                break;
            default:
                if (data[n].len < 0)
                    return 1;
                outcnt += 2 * sizeof(int) + data[n].len;
                if (buf)
                {
                    *(int *) buf = data[n].len;
                    buf += sizeof(int);
                    if (data[n].len)
                        memcpy(buf, data[n].d.p, data[n].len);
                    buf += data[n].len;
                }
                break;
        }
    }
    return outcnt;
}

/*! /brief Unpack a received structure
 *
 *  This is the reverse of ngcs_pack() - given packed data representing a
 *  structure, it creates an array of ngcs_data structures representing the
 *  contents of the structure. If you receive data marked as type 
 *  NGCS_TYPE_STRUCT, you should be able to unpack it using this function.
 *
 *  Note that this function allocates memory, which you must call
 *  ngcs_free_unpack(*res) to free once you're done. However, you must *not*
 *  do so if ngcs_unpack() fails (ie. if return value is < 0).
 *
 *  /param data the packed data representing the structure
 *  /param len the length of data, in bytes
 *  /param res a location where a pointer to an array of ngcs_data structures,
 *             representing the result of unpacking the data, is returned.
 *  /return the number of items unpacked, or -1 on error
 */
int ngcs_unpack(const char *data, int len, ngcs_data ** res)
{
    const char *d = data;
    int l = len;
    int cnt = 0;
    while (l >= 2 * (int) sizeof(int))
    {
        int head[2];
        memcpy(head, d, 2 * sizeof(int));
        if (head[1] < 0)
            return -1;
        d += 2 * sizeof(int) + head[1];
        l -= 2 * sizeof(int) + head[1];
        cnt++;
    }
    d = data;
    l = len;
    *res = malloc(cnt * sizeof(ngcs_data));
    cnt = 0;
    while (l >= 2 * (int) sizeof(int))
    {
        int head[2];
        memcpy(head, d, 2 * sizeof(int));
        (*res)[cnt].type = head[0];
        (*res)[cnt].len = head[1];
        if (head[1] < 0)
        {
            free(*res);
            *res = NULL;
            return -1;
        }
        d += 2 * sizeof(int);
        l -= 2 * sizeof(int);
        if (l < (*res)[cnt].len)
        {
            free(*res);
            *res = NULL;
            return -1;
        }
        switch ((*res)[cnt].type)
        {
            case NGCS_TYPE_INT:
                if ((*res)[cnt].len != sizeof(int))
                {
                    free(*res);
                    *res = NULL;
                    return -1;
                }
                (*res)[cnt].d.i = *(int *) d;
                break;
            case NGCS_TYPE_LONG:
                if ((*res)[cnt].len != sizeof(long))
                {
                    free(*res);
                    *res = NULL;
                    return -1;
                }
                (*res)[cnt].d.i = *(int *) d;
                break;
            case NGCS_TYPE_STRING:
                (*res)[cnt].d.s = malloc((*res)[cnt].len + 1);
                if (!(*res)[cnt].d.s)
                {
                    free(*res);
                    *res = NULL;
                    return -1;
                }
                memcpy((*res)[cnt].d.s, d, (*res)[cnt].len);
                (*res)[cnt].d.s[(*res)[cnt].len] = 0;
                break;
            default:
                (*res)[cnt].d.p = malloc((*res)[cnt].len);
                if (!(*res)[cnt].d.p)
                {
                    free(*res);
                    *res = NULL;
                    return -1;
                }
                memcpy((*res)[cnt].d.p, d, (*res)[cnt].len);
                break;
        }
        d += (*res)[cnt].len;
        l -= (*res)[cnt].len;
        cnt++;
    }
    return cnt;
}


/*! \brief Frees the memory allocated by ngcs_unpack()
 *  
 *  Once you're done with the results of ngcs_unpack(), you should
 *  call this method to free the memory allocated by it.
 *
 * /param len the number of data items unpacked (that is, the return value of ngcs_unpack())
 * /param res the array of ngcs_data structures returned by ngcs_unpack in its *res parameter
 */ 
void ngcs_free_unpack(int len, ngcs_data * res)
{
    int n;

    for (n = 0; n < len; n++)
    {
        switch (res[n].type)
        {
            case NGCS_TYPE_INT:
            case NGCS_TYPE_LONG:
                break;
            case NGCS_TYPE_STRING:
                if (res[n].d.s)
                    free(res[n].d.s);
                break;
            default:
                if (res[n].d.p)
                    free(res[n].d.p);
                break;
        }
    }
    free(res);
}

int ngcs_sendall(int sock, const void *buf, int len)
{
    int ret;

    while (len > 0)
    {
        ret = send(sock, buf, len, 0);
        if (ret <= 0)
        {
            return 1;
        }
        else
        {
            buf += ret;
            len -= ret;
        }
    }
    return 0;
}

int ngcs_recvall(int sock, void *buf, int len)
{
    int ret;

    while (len > 0)
    {
        ret = recv(sock, buf, len, 0);
        if (ret <= 0)
        {
            return 1;
        }
        else
        {
            buf += ret;
            len -= ret;
        }
    }
    return 0;
}
